/****************************************************************************
AirStick Driver for iMote2

Author: Brian Mayton <bmayton@cs.washington.edu>
Version: 0.1
****************************************************************************/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/i2c.h>
#include <linux/cdev.h>
#include <linux/poll.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <asm-arm/arch-pxa/pxa-regs.h>

#include "airstick.h"

MODULE_AUTHOR("Brian Mayton");
MODULE_DESCRIPTION("A driver for the AirStick board");
MODULE_VERSION("0.1");

dev_t dev_id;
struct cdev *cdev = NULL;

ssize_t airstick_read(struct file *, char __user *, size_t, loff_t *);
int airstick_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
int airstick_attach_adapter(struct i2c_adapter *adap);
int airstick_detach_client(struct i2c_client *client);

static struct file_operations airstick_fops = {
	.owner = THIS_MODULE,
	.ioctl = airstick_ioctl,
	.read = airstick_read,
};

static struct i2c_driver airstick_driver = {
	.owner = THIS_MODULE,
	.name = "airstick i2c Driver",
	.flags = I2C_DF_NOTIFY,
	.attach_adapter = airstick_attach_adapter,
	.detach_client = airstick_detach_client,
};

static struct i2c_client airstick_client = {
	.flags = 0,
	.addr = 0x42,
	.driver = &airstick_driver,
	.name = "AirStick i2c client",
};

int airstick_attach_adapter(struct i2c_adapter *adap) {
	unsigned char data[10];
	unsigned char id[5];
	unsigned char var[5];
	int ver;
	
	memset(data, 0, 10);
	
	printk("airstick_attach_adapter called\n");
	airstick_client.adapter = adap;
	i2c_attach_client(&airstick_client);
		
	data[0] = 0x00;
	i2c_master_send(&airstick_client, data, 1);
	i2c_master_recv(&airstick_client, data, 9);
		
	if(data[0] == 0x00) {
		i2c_detach_client(&airstick_client);
		printk("No AirStick on this adapter\n");
		return 1;
	}
	
	memcpy(id, data, 4);
	id[4] = 0;
	memcpy(var, data+4, 4);
	var[4] = 0;
	ver = data[8];
	
	printk("AirStick: %s variant %s, version %d\n", id, var, ver);
	
	msleep(100);
	
	return 0;
}

int airstick_detach_client(struct i2c_client *client) {
	printk("airstick_detach_client called\n");
	i2c_detach_client(client);
	printk("airstick_detach_client finished\n");
	return 0;
}

ssize_t airstick_read(struct file *filp, char __user *user_data, size_t count, loff_t *offp) {
	unsigned char data[8];	
	if(count > 8) count = 8;
	
	data[0] = 0x20;
	i2c_master_send(&airstick_client, data, 1);
	i2c_master_recv(&airstick_client, data, 8);
	
	copy_to_user(user_data, data, count);
	
	return count;
}

int airstick_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg) {
	unsigned char data[4];

	printk("airstick_ioctl called\n");
	
	switch(cmd) {
		case AIRSTICK_GET_STATUS:
			data[0] = 0x10;
			i2c_master_send(&airstick_client, data, 1);
			i2c_master_recv(&airstick_client, data, 1);
			return data[0];
		case AIRSTICK_POWER_ON:
			data[0] = 0x10;
			data[1] = 0x01;
			i2c_master_send(&airstick_client, data, 2);
			return 0;
		case AIRSTICK_POWER_OFF:
			data[0] = 0x10;
			data[1] = 0x00;
			i2c_master_send(&airstick_client, data, 2);
			return 0;
		case AIRSTICK_SET_NSAMPS:
			data[0] = 0x80;
			data[1] = arg & 0xFF;
			i2c_master_send(&airstick_client, data, 2);
			return 0;
	}
	
	return -ENOTTY;
}


static int __init airstick_init(void) {
	int result;
	printk("airstick_init called\n");

	
	// Allocate a major device number for this driver
	result = alloc_chrdev_region(&dev_id, 0, 1, "airstick");
	if (result < 0)
		return result;

	// Allocate a character device
	cdev = cdev_alloc();
	if (!cdev) {
		printk(KERN_WARNING "airstick: unable to allocate character device\n");
		unregister_chrdev_region(dev_id, 1);
		return -EAGAIN;
	}

	// Set the owner and file operations of this new character device
	cdev->owner = THIS_MODULE;
	cdev->ops = &airstick_fops;

	// Register the character device
	// Note at this point the device is live!
	result = cdev_add(cdev, dev_id, 1);
	if (result < 0) {
		printk(KERN_WARNING "airstick: unable to add character device\n");
		cdev_del(cdev);
		unregister_chrdev_region(dev_id, 1);
		return result;
	}
	
	return i2c_add_driver(&airstick_driver);
}

static void __exit airstick_exit(void) {
	printk("airstick_exit called\n");
	i2c_del_driver(&airstick_driver);
	
	// Free the character device
	cdev_del(cdev);
	// Unregister the major device number
	unregister_chrdev_region(dev_id, 1);
	
	printk("airstick_exit finished\n");
}

module_init(airstick_init);
module_exit(airstick_exit);
